﻿#include "waylandcapture.h"
#include <QtConcurrentRun>

#ifdef WAYLAND

#include "NemoSdkAdaptor.h"

WaylandCapture::WaylandCapture(QObject* parent)
    : QObject(parent)
    ,m_connectionThread(new QThread(this))
    ,m_connectionThreadObject(new ConnectionThread())
{
}

WaylandCapture::~WaylandCapture()
{
    if (m_remoteAccessManager) {
        if (m_buffer.mapped) {
            munmap(m_buffer.mapped_data, m_buffer.stride * m_buffer.height);
        }

        if (m_buffer.fd > 0)
        {
            close(m_buffer.fd);
        }

        m_remoteAccessManager->release();
        m_remoteAccessManager->destroy();
        m_remoteAccessManager = nullptr;
    }

    if (m_connectionThread) {
        if (m_connectionThread->isRunning()) {
            m_connectionThread->quit();
            m_connectionThread->wait();
        }

        m_connectionThreadObject->deleteLater();
    }
}

void WaylandCapture::startContentCapture()
{
    if (!_isFirstStart) {
        connect(m_connectionThreadObject, &ConnectionThread::connected, this, [this] {
            m_eventQueue = new EventQueue(this);
            m_eventQueue->setup(m_connectionThreadObject);

            registry = new Registry(this);
            setupRegistry(registry);
        }, Qt::QueuedConnection);
        m_connectionThreadObject->moveToThread(m_connectionThread);
        m_connectionThread->start();
        m_connectionThreadObject->initConnection();
        _isPause = false;
        _isFirstStart = true;
    }
    _isStopGrapScreen = false;
    _isPause = false;
}

void WaylandCapture::stopContentCapture()
{
    _isStopGrapScreen = true;
}

void WaylandCapture::pauseContentCapture() {
    _isPause = true;
}

void WaylandCapture::restartContentCapture() {
    _isPause = false;
}

void WaylandCapture::processBuffer(const Output* m_output,  RemoteBuffer* m_remoteBuffer)
{
    if (!m_remoteBuffer) {
        return;
    }

    if (m_buffer.mapped) {
        munmap(m_buffer.mapped_data, m_buffer.stride * m_buffer.height);
    }

    if (m_buffer.fd > 0)
        close(m_buffer.fd);
    
    QScopedPointer<const RemoteBuffer> guard(m_remoteBuffer);
    auto dma_fd = m_remoteBuffer->fd();

    // 非法的
    if (!dma_fd > 0)
        return;

    quint32 width = m_remoteBuffer->width();
    quint32 height = m_remoteBuffer->height();
    quint32 stride = m_remoteBuffer->stride();

    m_buffer.remoteBuffer = m_remoteBuffer;
    m_buffer.fd = dma_fd;
    m_buffer.width = width;
    m_buffer.height = height;
    m_buffer.stride = stride;

    unsigned char* mapData = static_cast<unsigned char*>(mmap(nullptr, stride * height, PROT_READ, MAP_SHARED, dma_fd, 0));

    if (MAP_FAILED == mapData) {
        qDebug() << "dma fd " << dma_fd << " mmap failed - ";
    } else {
        m_buffer.mapped_data = mapData;
        m_buffer.mapped = true;
        size_t bufLen = stride * height;
        _bufLen = bufLen;
        if (_rgbBuffer.get() == nullptr) {
            _rgbBuffer = std::shared_ptr<uint8_t>(new uint8_t[bufLen]);
        }

        memcpy(_rgbBuffer.get(), m_buffer.mapped_data, bufLen);

        _width = static_cast<int>(width);
        _height = static_cast<int>(height);
        _stride = static_cast<int>(stride);
        QtConcurrent::run([=]() {
            NemoSdkAdaptor::getInstance()->putContentData(_rgbBuffer.get(), bufLen, _width, _height, _stride);
        });
    }
}

void WaylandCapture::setupRegistry(Registry* registry)
{
    connect(registry, &Registry::compositorAnnounced, this, [this, registry](quint32 name, quint32 version) {
        m_compositor = registry->createCompositor(name, version, this);
    });

    connect(registry, &Registry::outputAnnounced, this, [this, registry] (quint32 name, quint32 version) {
        auto output = registry->createOutput(name, version, this);
    });

    connect(registry, &Registry::remoteAccessManagerAnnounced, this, [this, registry](quint32 name, quint32 version) {
        m_remoteAccessManager = registry->createRemoteAccessManager(name, version, this);
        connect(m_remoteAccessManager, &RemoteAccessManager::bufferReady, this, [this] (const void* output, const RemoteBuffer * m_remoteBuffer) {
            connect(m_remoteBuffer, &RemoteBuffer::parametersObtained, this, [this, m_remoteBuffer] {
                if (!_isStopGrapScreen && !_isPause) {
                    processBuffer(m_bufferReadyOutput, (RemoteBuffer*)m_remoteBuffer);
                }
            });
        });
    });

    connect(registry, &Registry::interfacesAnnounced, this, [this] {
        Q_ASSERT(m_compositor);
    });

    registry->setEventQueue(m_eventQueue);
    registry->create(m_connectionThreadObject);
    registry->setup();
}

#endif
